#include "common.h"
// #define DEBUG
//
#ifdef DEBUG
#define DEBUG_print(a) printf a;
#else
#define DEBUG_print(a)
#endif

FILE *DB;	// DB FH
struct stat DB_stat;  // DB file stat info
void *DB_map;	// DB mmap'd address
int  uid;	// Calling UID
int  euid;	// Effective UID
char *username;	// username

// Config file
const char *devzone_db = "/etc/devzone/user_zones.txt";

// Built to easily copy from config to struct :)
struct zone_info {
  char name[CONFIG_ZONE_LEN];
  char sep1;
  char username[CONFIG_USER_LEN];
  char sep2;
  char char_uid[CONFIG_UID_LEN];
  int uid;
};

/* Config file format:
 * 80 chars zonename padded with space:12 chars padded with space:5 chars padded with space\n
 * zone-a                                                                          :username    :uid     
 *
 * zones that are not reserved have a blank username and the uid field is blank or zero
 */

// Returns !0 on success, 0 on failure
// Also prints errors to stderr (bad style but quickly accomplished)
// TODO: check "username" against uid
int sanity_checks(int uid, int euid, char* username) {
  struct passwd *passwd_entry;

  // Check calling and effective UIDs, make sure we are effectively root
  if( euid != 0 ) {
    fprintf(stderr, "Error: not running with root permissions (euid==%d).\n", euid);
    return 0;
  }

  passwd_entry = getpwuid(uid);
  if( strcmp( passwd_entry->pw_name, username ) ) {
    fprintf(stderr, "Danger Will Robinson\n");
    return 0;
  }

// WARNING: enable debugging changes functionality of the program HERE
#ifdef DEBUG
  DEBUG_print(("Calling UID/Effective UID: %d/%d.\n", uid, euid));
#else // DEBUG
  // We don't want to actually BE root (unless we need to debug for some reason)
  if( uid == 0 ) {
    fprintf(stderr, "Error: Do not run directly as root (uid==%d).\n", euid);
    return 0;
  }
#endif // DEBUG

  return 1;
}

void init_DB() {

  // Open and lock db file
  DB = fopen(devzone_db, "r+");
  if( ! DB ) {
    perror("Could not open DB file");
    exit(-1);
  }
  lockf( fileno(DB), F_LOCK, 0 );
  // MMAP file for easy editing
  DB_map = mmap( NULL, 1048576, PROT_READ | PROT_WRITE, MAP_SHARED, fileno(DB), 0 );
  if( ! DB_map ) {
    perror("Could not mmap DB file");
    exit(-1);
  }
  fstat( fileno(DB), &DB_stat );

}

void close_DB() {

  // munmap
  lockf( fileno(DB), F_ULOCK, 0 );
  fclose(DB);

}

// exits on failure
void init() {
  DEBUG_print(("+init()\n"));

  // Ask system for calling/effective UIDs
  uid = getuid();
  euid = geteuid();
  // Ask the environment for a username (checked in sanity_checks())
  username = getenv("USER");

  if( ! sanity_checks(uid, euid, username) ) {
    exit(-1);
  }

  init_DB();

  DEBUG_print(("-init()\n"));
}

// exits on failure
void destroy() {
  close_DB();
}

// setup the zone info reference passed in based on
// the position of the record
void getZoneRecordByID(const int record_id, struct zone_info * ret) {
    int i; // iterate in a loop

    memcpy( ret, DB_map + CONFIG_LINE_LEN * record_id, CONFIG_LINE_LEN );

    // NULL out spaces to parse strings easily
    for( i=0;i<CONFIG_ZONE_LEN;i++)
      if( ret->name[i] == ' ' )
	ret->name[i] = '\0';
    for( i=0;i<CONFIG_USER_LEN;i++)
      if( ret->username[i] == ' ' )
	ret->username[i] = '\0';
    for( i=0;i<CONFIG_UID_LEN;i++)
      if( ret->char_uid[i] == ' ' )
	ret->char_uid[i] = '\0';

    // if we have a valid uid, parse it into a number
    if( ret->char_uid[0] != '\0' ) {
      ret->uid = atoi( ret->char_uid );
    } else {
      ret->uid = -1;
    }
}

// Let's be paranoid...
void updateRecordID(int recordID, struct zone_info *newInfo) {
  char newLine[CONFIG_LINE_LEN + 1]; // to be copied into the config
  int i;

  DEBUG_print(("-updateRecordID(%d, ...)\n", recordID));

  for( i = 0; i <= CONFIG_LINE_LEN; i++ )
    newLine[i] = ' ';

  memcpy( newLine, newInfo->name,     strlen(newInfo->name) );
  memcpy( newLine + CONFIG_ZONE_LEN + 1, newInfo->username, strlen(newInfo->username) );
  memcpy( newLine + CONFIG_ZONE_LEN + CONFIG_USER_LEN + 2, newInfo->char_uid, strlen(newInfo->char_uid) );

  for( i = 0; i <= CONFIG_LINE_LEN; i++ )
    if( newLine[i] == '\0' )
      newLine[i] = ' ';

  newLine[CONFIG_ZONE_LEN] = CONFIG_SEP;
  newLine[CONFIG_ZONE_LEN+1+CONFIG_USER_LEN] = CONFIG_SEP;
  newLine[CONFIG_LINE_LEN-1] = '\n';
  newLine[CONFIG_LINE_LEN +1] = '\0';
  DEBUG_print(("new DB line: %s", newLine));
  memcpy(DB_map + CONFIG_LINE_LEN * recordID, newLine, CONFIG_LINE_LEN);
  msync(DB_map + CONFIG_LINE_LEN * recordID, CONFIG_LINE_LEN, MS_ASYNC);

  DEBUG_print(("-updateRecordID(%d, ...)\n", recordID));
}

int getZoneRecordCount() {
  return DB_stat.st_size/CONFIG_LINE_LEN; 
}

char *getReservedZoneName() {
  char *ret_zonename = NULL;
  int   i = 0;
  struct zone_info currentZone;

  DEBUG_print(("+getReservedZoneName()\n"));

  for( i = 0; i < getZoneRecordCount(); i++ ) {
    getZoneRecordByID(i, &currentZone);

    DEBUG_print(( "%s - %s (%d)\n", currentZone.name, currentZone.username, currentZone.uid ));

    // Did we find the reserved zone name?
    if( currentZone.uid == uid ) {
      ret_zonename = (char *)malloc(sizeof(char)*(1+strlen(currentZone.name)));
      strcpy( ret_zonename, currentZone.name );

      DEBUG_print(("-getReservedZoneName(): %s\n", ret_zonename));

      return ret_zonename;
    }

  }

  DEBUG_print(("-getReservedZoneName(): NULL\n"));

  // We did not find the reserved zone
  return NULL;
}

char *getFreeZoneName() {
  char *ret_zonename = NULL;
  int   i = 0;
  struct zone_info currentZone;

  DEBUG_print(("+getFreeZoneName()\n"));

  for( i = 0; i < getZoneRecordCount(); i++ ) {
    getZoneRecordByID(i, &currentZone);

    DEBUG_print(( "%s - %s (%d)\n", currentZone.name, currentZone.username, currentZone.uid ));

    // Did we find the reserved zone name?
    if( currentZone.uid == -1 ) {
      ret_zonename = (char *)malloc(sizeof(char)*(1+strlen(currentZone.name)));
      strcpy( ret_zonename, currentZone.name );

      DEBUG_print(("-getFreeZoneName(): %s\n", ret_zonename));

      return ret_zonename;
    }
  }

  DEBUG_print(("-getFreeZoneName(): NULL\n"));

  return NULL;
}

// helper for things that want this.
char *getUsername() {
  return username;
}

char *reserveZone(char *target_zone_name) {
  char *ret_zonename = NULL;
  int   i = 0;
  struct zone_info currentZone;

  DEBUG_print(("+reserveZone()\n"));

  for( i = 0; i < getZoneRecordCount(); i++ ) {
    getZoneRecordByID(i, &currentZone);

    DEBUG_print(( "%s - %s (%d)\n", currentZone.name, currentZone.username, currentZone.uid ));

    if( strcmp( target_zone_name, currentZone.name ) != 0 ) {
      DEBUG_print(("no match...\n"));
      continue;
    }

    DEBUG_print(("Found db record (%d), update line in file...\n", i)); 
    DEBUG_print(("%s:%s:%d\n", target_zone_name, username, uid));

    // Setup zone with new info
    memcpy(currentZone.username, username, strlen(username));
    currentZone.uid = uid;
    sprintf(currentZone.char_uid, "%d", uid);

    updateRecordID(i, &currentZone);

    ret_zonename = (char *)malloc(sizeof(char)*(1+strlen(currentZone.name)));
    strcpy( ret_zonename, currentZone.name );
    DEBUG_print(("-reserveZone(): %s\n", ret_zonename));
    return ret_zonename;
  }

  DEBUG_print(("-reserveZone(): NULL\n"));
  return NULL;
}

char *freeZone(char *target_zone_name) {
  char *ret_zonename = NULL;
  int   i = 0;
  struct zone_info currentZone;

  DEBUG_print(("+freeZone()\n"));

  for( i = 0; i < getZoneRecordCount(); i++ ) {
    getZoneRecordByID(i, &currentZone);

    DEBUG_print(( "%s - %s (%d)\n", currentZone.name, currentZone.username, currentZone.uid ));

    if( strcmp( target_zone_name, currentZone.name ) != 0 ) {
      DEBUG_print(("no match...\n"));
      continue;
    }

    DEBUG_print(("Found db record (%d), update line in file...\n", i)); 
    DEBUG_print(("%s:%s:%d\n", target_zone_name, username, uid));

    // Setup zone with new info
    currentZone.username[0] = '\0';
    currentZone.uid = -1;
    currentZone.char_uid[0] = '\0';

    updateRecordID(i, &currentZone);

    ret_zonename = (char *)malloc(sizeof(char)*(1+strlen(currentZone.name)));
    strcpy( ret_zonename, currentZone.name );
    DEBUG_print(("-freeZone(): %s\n", ret_zonename));
    return ret_zonename;
  }

  DEBUG_print(("-freeZone(): NULL\n"));
  return NULL;
}
